"""Top-level package for fastscape_litho."""

__author__ = """Boris Gailleton"""
__email__ = 'boris.gailleton@gfz-potsdam.de'
__version__ = '0.0.1'


from .fastscape_litho import *
from .TA import *
from ._looper import *
from ._helper import *
